""""
Code-behind the scenes for the cli application.
"""
import os

from clokta.aws_cred_generator import AwsCredentialsGenerator
from clokta.common import Common
from clokta.okta_initiator import OktaInitiator
from clokta.clokta_configuration import CloktaConfiguration


class RoleAssumer(object):
    """ Core implementation of clokta """

    def __init__(self, profile):
        """
        :param profile: the name of the AWS profile the user wants to clokta into (e.g. pagebuilder)
        :type profile: str
        """
        self.profile = profile
        """folder to store files in"""
        self.data_dir = "~/.clokta/"
        if not os.path.exists(os.path.expanduser(self.data_dir)):
            os.mkdir(os.path.expanduser(self.data_dir))

    def assume_role(self, reset_default_role):
        """
        entry point for the cli tool
        :param reset_default_role: whether to reset whatever the default role is for this profile
        :type reset_default_role: bool
        """
        clokta_config_file = self.data_dir + "clokta.cfg"

        clokta_config = CloktaConfiguration(profile_name=self.profile,
                                            clokta_config_file=clokta_config_file)
        if reset_default_role:
            clokta_config.reset_default_role()

        # Attempt to initiate a connection using just cookies
        okta_initiator = OktaInitiator(data_dir=self.data_dir)
        result = okta_initiator.initiate_with_cookie(clokta_config)

        # If the cookie is expired or non-existent, INPUT_ERROR will be returned
        if result == OktaInitiator.Result.INPUT_ERROR:
            prompt_for_password = clokta_config.get('okta_password') is None
            mfas = []
            # Cookie didn't work.  Authenticate with Okta
            while result == OktaInitiator.Result.INPUT_ERROR:
                if prompt_for_password:
                    clokta_config.prompt_for(param_name='okta_password')
                result = okta_initiator.initiate_with_auth(clokta_config, mfas)
                if result == OktaInitiator.Result.INPUT_ERROR:
                    if prompt_for_password:
                        Common.dump_err("Failure.  Wrong password or misconfigured session.")
                    else:
                        Common.dump_err("Saved password may be out of date.")
                prompt_for_password = True

            if result == OktaInitiator.Result.NEED_MFA:
                done = False
                first_time = True
                while not done:
                    chosen_factor = clokta_config.determine_mfa_mechanism(mfas, force_prompt=not first_time)
                    need_otp = okta_initiator.initiate_mfa(factor=chosen_factor)
                    otp = clokta_config.determine_okta_onetimepassword(chosen_factor, first_time) if need_otp else None
                    result = okta_initiator.finalize_mfa(clokta_config=clokta_config, factor=chosen_factor, otp=otp)
                    done = result == OktaInitiator.Result.SUCCESS
                    first_time = False

        saml_assertion = okta_initiator.saml_assertion

        # We now have a SAML assertion and can generate a AWS Credentials
        aws_svc = AwsCredentialsGenerator(clokta_config=clokta_config,
                                          saml_assertion=saml_assertion,
                                          data_dir=self.data_dir)
        roles = aws_svc.get_roles()
        role = clokta_config.determine_role(roles)
        aws_svc.generate_creds(role)
        clokta_config.update_configuration()

        self.output_instructions(docker_file=aws_svc.docker_file, bash_file=aws_svc.bash_file)

    def output_instructions(self, docker_file, bash_file):
        if Common.get_output_format() == Common.quiet_out:
            Common.echo(
                message='export AWS_PROFILE={}'.format(self.profile),
                always_stdout=True
            )
        elif Common.get_output_format() == Common.long_out:
            Common.echo(
                message='\nAWS keys generated.\nTo use with docker-compose include\n' +
                        '\tenv_file:\n\t    - {}\n'.format(docker_file) +
                        'To use with shell scripts include\n\tsource {}\n'.format(bash_file) +
                        'to use in the current interactive shell run\n\texport AWS_PROFILE={}\n'.format(self.profile)
            )
        else:
            Common.echo(
                message='Add the "-i" flag for how to use credentials and override defaults or just run:\n\n' +
                        'export AWS_PROFILE={}\n'.format(self.profile)
            )



import click

from clokta.common import Common
from clokta.factors import Factors


class FactorChooser(object):
    """ Supports MFA source determination """

    def __init__(self, factors, factor_preference=None):
        """ Instance constructor """
        self.okta_factors = factors
        self.factor_preference = factor_preference

        self.cli_factors = self.__load_supported_factors()
        self.option_factors = self.__filter_unsupported_factors()

    def verify_only_factor(self, factor):
        """ Return the Okta MFA configuration provided it is a supported configuration """
        verified_factors = [
            opt for opt in self.option_factors
            if opt['provider'] == factor['provider'] and
            opt['factor_type'] == factor['factorType']
        ]
        if verified_factors:
            if Common.is_debug():
                msg = 'Using only available factor: {}'.format(verified_factors[0]['prompt'])
                Common.dump_out(message=msg)
            return factor

    def verify_preferred_factor(self):
        """ Return the Okta MFA configuration for the matching, supported configuration """
        preferred_factors = [
            opt for opt in self.option_factors
            if self.factor_preference == opt['prompt']
        ]
        if preferred_factors:
            if Common.is_debug():
                msg = 'Using preferred factor: {}'.format(self.factor_preference)
                Common.dump_out(message=msg)

            matching_okta_factor = [
                fact for fact in self.okta_factors
                if fact['provider'] == preferred_factors[0]['provider'] and
                fact['factorType'] == preferred_factors[0]['factor_type']
            ]

            return matching_okta_factor[0]

        else:
            msg = 'The MFA option \'{}\' in your configuration file is not available.\nAvailable options are {}'.format(
                self.factor_preference, [opt['prompt'] for opt in self.option_factors])
            Common.dump_err(message=msg)
            raise ValueError("Unexpected MFA option")  # TODO: Reprompt

    def choose_supported_factor(self):
        """ Give the user a choice from the intersection of configured and supported factors """
        index = 1
        for opt in self.option_factors:
            msg = '{index} - {prompt}'.format(index=index, prompt=opt['prompt'])
            Common.echo(message=msg, bold=True)
            index += 1

        raw_choice = None
        try:
            raw_choice = click.prompt('Choose a MFA type to use', type=int, err=Common.to_std_error())
            choice = raw_choice - 1
        except ValueError:
            Common.echo(message='Please select a valid option: you chose: {}'.format(raw_choice))
            return self.choose_supported_factor()

        if len(self.option_factors) > choice >= 0:
            pass
        else:
            Common.echo(message='Please select a valid option: you chose: {}'.format(raw_choice))
            return self.choose_supported_factor()

        chosen_option = self.option_factors[choice]
        matching_okta_factor = [
            fact for fact in self.okta_factors
            if fact['provider'] == chosen_option['provider'] and
            fact['factorType'] == chosen_option['factor_type']
        ]
        if Common.is_debug():
            Common.dump_out(message='Using chosen factor: {}'.format(chosen_option['prompt']))

        return matching_okta_factor[0]

    def __load_supported_factors(self):
        return Factors.mfa_providers()

    def __filter_unsupported_factors(self):
        factor_intersection = []
        for cli in self.cli_factors:
            for okta in self.okta_factors:
                if cli['provider'] == okta['provider'] and cli['factor_type'] == okta['factorType']:
                        factor_intersection.append(cli)
                        okta['clokta_id'] = cli['prompt']
        return factor_intersection



import enum


class ConfigParameter:
    """
    This is a structure to hold all the information about a single configured parameter used by Clokta
    (e.g. okta_username)
    """

    class SaveTo(enum.Enum):
        """ Enumeration for to where to save configuration parameter's value """
        DONT_SAVE = 0  # Don't save this parameter anywhere
        DEFAULT = 1  # Save to the default section of the clokta config file
        PROFILE = 2  # Save to the profile section of the clokta config file
        KEYRING = 3  # For passwords.  Save to the OS keyring

    def __init__(self, name, value=None, required=False, save_to=SaveTo.DONT_SAVE,
                 secret=False, default_value=None, prompt=None, param_type=str):
        """
        :param name: the name of the parameter (e.g. okta_username)
        :type name: str
        :param value: the value it has been set to (e.g. doej)
        :type value: str
        :param required: whether this value is required for a clokta login and the user should be
            prompted for a value if none is known.  okta_username is required.  okta_aws_role_to_assume is
            often deduced in the process though may be prompted for later
        :type required: boolean
        :param save_to: where to persistently store this parameters value so we remember it for
            next time.  Options are SAVE_TO_DEFAULT to be stored in the default section of the
            clokta.cfg file, SAVE_TO_PROFILE to be stored in the clokta.cfg file in the section for
            this specific profile, and SAVE_TO_KEYRING for storing passwords
        :type save_to: SaveTo
        :param secret: whether this is a password and should be prompted without echoing
        :type secret: bool
        :param default_value: a default value to use if none is specified
        :type default_value: str or bool
        :param param_type: the type of the parameter.  Usally they are str, but bool is also supported
        :type param_type: type
        :param prompt: the string to use to prompt the user for this parameter
        :type prompt: str
        """
        self.name = name
        self.value = value
        self.required = required
        self.save_to = save_to
        self.secret = secret
        self.default_value = default_value
        self.prompt = prompt
        self.param_type = param_type




import base64
import getpass
import sys

import click
import configparser
import enum
import json
import keyring
import os
import re

from clokta.common import Common
from clokta.config_parameter import ConfigParameter
from clokta.factor_chooser import FactorChooser
from clokta.role_chooser import RoleChooser


class OutputFormat(enum.Enum):
    """ Enumeration for credentials output format codes """
    ShellScript = 'SHELL'
    Profile = 'PROFILE'
    All = 'ALL'


class CloktaConfiguration(object):
    """
    This handles collecting and storing the configuration paramaters needed to clokta into a system
    It encapsulates reading and storing values from the clokta.cfg file, reading and storing passwords from
    keyring, and prompting the user for values.
    """

    KEYCHAIN_PATTERN = "clokta.{param_name}"  # The key in the keychain to use when storing a param

    def __init__(
        self,
        profile_name,
        clokta_config_file,
        profiles_location='~/.aws/credentials'
    ):
        """ Instance constructor """
        self.profile_name = profile_name
        self.profiles_location = os.path.expanduser(profiles_location)
        self.clokta_config_file = os.path.expanduser(clokta_config_file)
        self.param_list = self.__define_parameters()  # type: [ConfigParameter]
        self.parameters = {p.name: p for p in self.param_list}
        self.displayed_python2_warning = False
        self.__initialize_configuration()

    def get(self, parameter_name):
        """
        Get the current value of a configuration parameter
        :param parameter_name: the name of the parameter
        :type parameter_name: str
        :return: the value of the parameter or None if not present
        """
        return self.parameters[parameter_name].value if parameter_name in self.parameters else None

    def update_configuration(self):
        """
        Write the current version of the configuration to the clokta.cfg file
        """
        clokta_cfg_file = configparser.ConfigParser()
        clokta_cfg_file.read(self.clokta_config_file)
        if not clokta_cfg_file.has_section(self.profile_name):
            clokta_cfg_file.add_section(self.profile_name)

        for param in self.param_list:
            if param.value:
                if param.save_to == ConfigParameter.SaveTo.DEFAULT:
                    clokta_cfg_file.set('DEFAULT', param.name, param.value)
                elif param.save_to == ConfigParameter.SaveTo.PROFILE:
                    clokta_cfg_file.set(self.profile_name, param.name, param.value)
                elif param.save_to == ConfigParameter.SaveTo.KEYRING:
                    self.__save_to_keyring(param.name, param.value)

        if Common.is_debug():
            Common.dump_out(
                message='Re-writing configuration file {}'.format(self.clokta_config_file)
            )
        self.__write_config(
            path_to_file=self.clokta_config_file,
            parser=clokta_cfg_file
        )

    def reset_default_role(self):
        # Clear it from the clokta.cfg file
        clokta_cfg_file = configparser.ConfigParser()
        clokta_cfg_file.read(self.clokta_config_file)
        clokta_cfg_file.remove_option(self.profile_name, 'okta_aws_role_to_assume')
        self.__write_config(
            path_to_file=self.clokta_config_file,
            parser=clokta_cfg_file
        )

        # And reset it
        self.parameters['okta_aws_role_to_assume'].value = None

    def __define_parameters(self):
        """
        Hard coded definition of all possible parameters
        :return: parameters without values
        :rtype: [ConfigParameter]
        """

        parameters = [
            ConfigParameter(
                name='okta_username',
                required=True,
                save_to=ConfigParameter.SaveTo.DEFAULT
            ),
            ConfigParameter(
                name='multifactor_preference',
                save_to=ConfigParameter.SaveTo.DEFAULT,
                default_value='Okta Verify with Push'
            ),
            ConfigParameter(
                name='okta_aws_app_url',
                required=True,
                save_to=ConfigParameter.SaveTo.PROFILE
            ),
            ConfigParameter(
                name='okta_password',
                secret=True
            ),
            ConfigParameter(
                name='save_password_in_keychain',
                required=True,
                save_to=ConfigParameter.SaveTo.DEFAULT,
                prompt='Would you like clokta to remember your password? [Yes]',
                default_value=True,
                param_type=bool
            ),
            ConfigParameter(
                name='okta_aws_role_to_assume'
            ),
            ConfigParameter(
                name='okta_onetimepassword_secret',
                secret=True
            ),
            ConfigParameter(
                # aws_account_number is not really an input parameter, but
                # something we deduce during login and wanted to save in the clokta.cfg
                name='aws_account_number',
                save_to=ConfigParameter.SaveTo.PROFILE
            )
        ]
        return parameters


    def __check_url(self, url):
        """
        Verify that the URL passed in is an Okta App URL to an AWS app.  Will raise an exception
        if url is invalid
        :return: the valid URL with any query parameters stripped off
        :rtype: str
        """
        # The URL should be of the form (though query parameter "fromHome" is optional)
        # https://<YOUR_COMPANY>.okta.com/home/amazon_aws/<UNIQUE_ID>/<APP_ID>?fromHome=true
        pattern = re.compile('https://[^.]+.okta.com/home/amazon_aws/[^/]+/[0-9]+\\??')
        match = pattern.match(url)
        if match is None:
            Common.dump_err(
                "Invalid App URL.  URL usually of the form https://xxxxxxxx.okta.com/home/amazon_aws/...", 6
            )
            raise ValueError("Invalid URL")
        else:
            # Strip off any query parameters
            paramIndex = url.find('?');
            if paramIndex >= 0:
                url = url[:paramIndex]

        return url;


    def __initialize_configuration(self):
        """
        Load config file, both the desired section and the default section
        :return: the parameter list
        :rtype: dict
        """
        clokta_cfg_file = configparser.ConfigParser()
        clokta_cfg_file.read(self.clokta_config_file)

        # Make sure we have the bare minimum in the config file.  The DEFAULT section and the app URL.
        if not clokta_cfg_file['DEFAULT']:
            clokta_cfg_file['DEFAULT'] = {
            }
        if not clokta_cfg_file.has_section(self.profile_name):
            msg = 'No profile "{}" in clokta.cfg, but enter the information and clokta will create a profile.\n' + \
                  'Copy the link from the Okta App'
            app_url = click.prompt(text=msg.format(self.profile_name), type=str, err=Common.to_std_error()).strip()
            app_url = self.__check_url(app_url)
            clokta_cfg_file.add_section(self.profile_name)
            clokta_cfg_file.set(self.profile_name, 'okta_aws_app_url', app_url)

        config_section = clokta_cfg_file[self.profile_name]
        self.__load_parameters(config_section)
        if self.get('save_password_in_keychain') == 'True':
            self.parameters['okta_password'].save_to = ConfigParameter.SaveTo.KEYRING

    def __prompt_for(self, param):
        prompt = param.prompt if param.prompt else 'Enter a value for {}'.format(param.name)
        if param.secret:
            field_value = getpass.getpass(prompt=prompt+":")
        else:
            field_value = click.prompt(text=prompt,
                                       type=param.param_type,
                                       err=Common.to_std_error(),
                                       default=param.default_value,
                                       show_default=not param.prompt)
        return field_value if param.param_type==str else str(field_value)

    def apply_credentials(self, credentials):
        """ Save a set of temporary credentials """
        if Common.is_debug():
            msg = json.dumps(obj=credentials, default=Common.json_serial, indent=4)
            Common.dump_out(message=msg)

        parser = configparser.ConfigParser()
        parser.read(self.profiles_location)

        if self.profile_name not in parser.sections():
            if Common.is_debug():
                Common.dump_out(
                    message='Adding profile section {}'.format(self.profile_name)
                )
            parser.add_section(self.profile_name)

        creds = credentials['Credentials']
        parser[self.profile_name]['AWS_ACCESS_KEY_ID'] = creds['AccessKeyId']
        parser[self.profile_name]['AWS_SECRET_ACCESS_KEY'] = creds['SecretAccessKey']

        if 'AWS_SESSION_TOKEN' in parser[self.profile_name]:
            del parser[self.profile_name]['AWS_SESSION_TOKEN']

        if 'SessionToken' in creds:
            parser[self.profile_name]['AWS_SESSION_TOKEN'] = creds['SessionToken']

        if Common.is_debug():
            Common.dump_out(
                message='Re-writing credentials file {}'.format(self.profiles_location)
            )

        self.__write_config(
            path_to_file=self.profiles_location,
            parser=parser
        )

    def __write_config(self, path_to_file, parser):
        """ Write config to file """
        self.__backup_file(path_to_file=path_to_file)

        if not os.path.exists(os.path.dirname(path_to_file)):
            os.makedirs(os.path.dirname(path_to_file))

        with open(path_to_file, 'w') as file:
            parser.write(file)

    def __backup_file(self, path_to_file):
        """ Back up config """
        backup_location = os.path.expanduser(
            '{}.bak'.format(
                path_to_file
            )
        )
        if os.path.isfile(path=path_to_file):
            with open(path_to_file, 'r') as file:
                contents = file.read()

            with open(backup_location, 'w') as bak_file:
                bak_file.write(contents)

    def determine_mfa_mechanism(self, mfas, force_prompt):
        """
        Determine which of the passed in MFA mechanisms to use.  This may be specified
        by the configuration's 'multifactor_preference' if not prompt the user and put
        in the configuration
        :param mfas: possible mechanisms to use for MFA
        :type mfas: List[dict]
        :param force_prompt: whether to ignore any preconfigured default and prompt the user
        :type force_prompt: bool
        :return: the chosen MFA mechanism to use
        :rtype: dict
        """

        factor_preference = self.get('multifactor_preference')
        fact_chooser = FactorChooser(
            factors=mfas,
            factor_preference=factor_preference
        )

        chosen_factor = None
        # Is there only one legitimate MFA option available?
        if len(mfas) == 1:
            chosen_factor = fact_chooser.verify_only_factor(factor=mfas[0])

        if not chosen_factor:
            # Has the user pre-selected a legitimate factor?
            if factor_preference and not force_prompt:
                chosen_factor = fact_chooser.verify_preferred_factor()

        if not chosen_factor:
            chosen_factor = fact_chooser.choose_supported_factor()
            if chosen_factor:
                self.parameters['multifactor_preference'].value = chosen_factor['clokta_id']

        return chosen_factor

    def determine_role(self, possible_roles):
        """
        Determine which of several possible roles to assume by looking first in the config for a default role,
        and second by prompting the user.
        :param possible_roles: list of possible roles to assume
        :type possible_roles: List[AwsRole]
        :return: the role chosen
        :rtype: AwsRole
        """
        role_chooser = RoleChooser(
            possible_roles=possible_roles,
            role_preference=self.get('okta_aws_role_to_assume')
        )
        chosen_role, make_default = role_chooser.choose_role()
        self.parameters['okta_aws_role_to_assume'].value = chosen_role.role_arn
        if make_default:
            self.parameters['okta_aws_role_to_assume'].save_to = ConfigParameter.SaveTo.PROFILE
        # We also grab the account number out of the role ARN
        self.parameters['aws_account_number'].value = chosen_role.account
        return chosen_role

    def prompt_for(self, param_name):
        """
        Prompt the user for the parameter and store it in the configuration
        :param param_name: the name of the parameter
        :type param_name: str
        """
        param_to_prompt_for = self.parameters[param_name]
        param_to_prompt_for.value = self.__prompt_for(param_to_prompt_for)

    def determine_okta_onetimepassword(self, factor, first_time):
        """
        Get the one time password, which may be in one password or
        may need to be prompted for
        :param factor: the mfa mechanism being used.  Holds a user friendly label for identifying which mechanism.
        :type factor: dict
        :param first_time: whether this is the first time this run determining one time password
            on subsequent attempts we don't try to get it with an otp secret because that
            obviously didn't work the first time
        :return: the Okta one time password
        :rtype: string
        """

        otp_value = None
        if self.get('okta_onetimepassword_secret'):
            if not first_time:
                Common.dump_err("OTP generator created incorrect OTP")
            else:
                try:
                    # noinspection PyUnresolvedReferences
                    import onetimepass as otp
                except ImportError:
                    msg = 'okta_onetimepassword_secret provided in config but "onetimepass" is not installed. ' + \
                          'run: pip install onetimepass'
                    Common.dump_err(message=msg)
                    raise ValueError("Illegal configuration")
                otp_value = otp.get_totp(self.get('okta_onetimepassword_secret'))

        if not otp_value:
            otp_value = click.prompt(
                text='Enter your {} one time password'.format(factor['clokta_id']),
                type=str,
                err=Common.to_std_error(),
                default=''
            )
        return otp_value

    def __load_parameters(self, config_section):
        """
        For each parameter this will look first in the OS environment, then in the
        config file (which will look in both the section and in the DEFAULT section),
        then in the keychain (for secrets only) and then
        if still not found and the attribute is required, will prompt the user.
        :param config_section: section of the clokta.cfg file that represents the profile that we will
        login to though queries on this will also look in the DEFAULT section
        :type config_section:
        :return: a map of attributes that define the clokta login, e.g.
            {"okta_username": "doej", "multifactor_preference": "Google Authenticator", ...}
        :rtype: map[string, string]
        """
        debug_msg = 'Configuration:\n'
        for param in self.param_list:
            from_env = os.getenv(key=param.name, default=-1)
            if from_env != -1:
                # If defined in environment, use that first
                param.value = from_env
            elif param.name in config_section and config_section[param.name]:
                # If defined in the config file, make sure it's not a secret, otherwise use it
                if param.secret:
                    Common.dump_err(
                        message='Invalid configuration.  {} should never be defined in clokta.cfg.'.format(param.name))
                    raise ValueError("Illegal configuration")
                if param.param_type == bool:
                    param.value = self.__validate_bool(config_section[param.name], param.name)
                else:
                    param.value = config_section[param.name]
            elif param.secret:
                param.value = self.__read_from_keyring(param.name)

            if not param.value and param.required:
                # We need it.  Prompt for it.
                param.value = self.__prompt_for(param)
            debug_msg += '     {}={}'.format(param.name, param.value if not param.secret else 'xxxxxxxx')

        if Common.is_debug():
            Common.dump_out(message=debug_msg)

    def __read_from_keyring(self, param_name):
        """
        Read a secret from the OS keychain
        :param param_name: the name of the parameter
        :type param_name: str
        :return: the value or None if not found or otherwise could not get value
        :rtype: str
        """
        param_value = None
        if sys.version_info > (3, 0):
            system = CloktaConfiguration.KEYCHAIN_PATTERN.format(param_name=param_name)
            user = self.get('okta_username')
            try:
                obfuscated = keyring.get_password(system, user)
                param_value = self.__deobfuscate(obfuscated, user)
            except Exception as e:
                fail_msg = str(e)
                if fail_msg.find('Security Auth Failure') >= 0:
                    Common.dump_err('WARNING: Denied access to password in keychain.  ' +
                                    'If prompted by keychain, allow access.\n' +
                                    'You may need to reboot your machine before keychain will prompt again.')
                else:
                    Common.dump_err('WARNING: Could not read password from keychain: {}'.format(e))
        else:
            if not self.displayed_python2_warning:
                Common.dump_err("Cannot store password in keychain.  Upgrade to Python 3 to use keychain.")
                self.displayed_python2_warning = True
        return param_value

    def __save_to_keyring(self, param_name, param_value):
        """
        Save a secret to the keychain.  Obfuscate the secret first.
        :param param_name: the name of the parameter
        :type param_name: str
        :param param_value: the secret to save
        :type param_value: str
        """
        if sys.version_info > (3, 0):
            try:
                system = CloktaConfiguration.KEYCHAIN_PATTERN.format(param_name=param_name)
                user = self.get('okta_username')
                password = self.__obfuscate(param_value, user)
                keyring.set_password(system, user, password)
            except Exception as e:
                fail_msg = str(e)
                if fail_msg.find('Security Auth Failure') >= 0:
                    Common.dump_err(fail_msg)
                    Common.dump_err('WARNING: Denied access to password in keychain.  ' +
                                    'If prompted by keychain, allow access.\n' +
                                    'You may need to reboot your machine before keychain will prompt again.')
                else:
                    Common.dump_err('WARNING: Could not save password to keychain: {}'.format(e))

    def __obfuscate(self, secret, salt):
        """
        A real simple obfuscation so passwords don't appear in the keychain as plaintext
        This is not encryption and can easily be cracked or reverse engineered
        Counting on the keyring's security to protect the password
        :param secret: a secret to obfuscate
        :type secret: str
        :param salt: a phrase to use to obfuscate.  Will be used again to deobfuscate.
        :type salt: str
        :return: a obfuscated version that can be turned back into the secret with __deobfuscate
        :rtype: str
        """
        if not secret:
            return None

        enc = []
        for i in range(len(secret)):
            salt_c = salt[i % len(salt)]
            enc_c = chr((ord(secret[i]) + ord(salt_c)) % 256)
            enc.append(enc_c)
        encoded = base64.urlsafe_b64encode("".join(enc).encode()).decode()

        return encoded

    def __deobfuscate(self, obfuscated, salt):
        """
        Turn a string that was obfuscated with __obfuscate back into the original
        :param obfuscated: obfuscated string
        :type obfuscated: str
        :param salt: the phrase used to obfuscate the secret
        :type salt: str
        :return: the original string
        :rtype: str
        """
        if not obfuscated:
            return None

        dec = []

        enc = base64.urlsafe_b64decode(obfuscated).decode()

        for i in range(len(enc)):
            salt_c = salt[i % len(salt)]
            dec_c = chr((256 + ord(enc[i]) - ord(salt_c)) % 256)
            dec.append(dec_c)
        return "".join(dec)

    def __validate_bool(self, value, name="parameter"):
        """
        Verify a string is either True or False.
        If it is not, print an error.
        :param value: the string value
        :type value: str
        :param name: the name or context of the value to be used in an error message
        :type name: str
        :return: "True" if the string is some form of true,  "False" for any other case
        :rtype: str
        """
        if value.strip().lower() == "true":
            return "True"
        elif value.strip().lower() == "false":
            return "False"
        else:
            Common.dump_err('{} configured with value "{}" when only True or False is valid.'.format(name, value))
            return "False"

    @classmethod
    def dump_account_numbers(cls, clokta_config_file):
        clokta_cfg_file = configparser.ConfigParser()
        clokta_cfg_file.read(os.path.expanduser(clokta_config_file))
        section_names = clokta_cfg_file.sections()
        for section_name in section_names:
            if clokta_cfg_file.has_option(section=section_name, option='aws_account_number'):
                acct_num = clokta_cfg_file.get(section=section_name, option='aws_account_number')
                if acct_num:
                    Common.echo("{name} = {number}".format(name=section_name, number=acct_num))



import os

from clokta.clokta_configuration import CloktaConfiguration
import json
import pickle
import requests
import time
from enum import Enum

from bs4 import BeautifulSoup

from clokta.common import Common


class OktaInitiator:

    class Result(Enum):
        INPUT_ERROR = 1
        NEED_MFA = 2
        SUCCESS = 3

    def __init__(self, data_dir):
        """
        This class initiates connections with Okta.
        Once you attempt to initiate a connection this object has three states
        1) Failed - you can query the failed state or try again
        2) Waiting on MFA - you can submit MFA credentials
        3) Succeeded - you can retrieve the SAML token

        :param data_dir: the directory to store the cookies file
        :type data_dir: str
        """
        self.data_dir = os.path.expanduser(data_dir)
        self.saml_assertion = None  # type: str
        self.session_token = None  # type: str
        self.intermediate_state_token = None  # type: str
        self.factors = []  # type: [dict]

    def get_saml_assertion(self):
        """
        :return: the SAML token, or None if not successfully retrieved
        :rtype: str
        """
        return self.saml_assertion

    def initiate_with_cookie(self, clokta_config):
        """
        Request a SAML token from Okta without authenticating but relying on valid Okta cookie
        from a previous interaction.  If succesful state will be SUCCESS and you can get
        the SAML token.  If unsuccesful state will be FAIL.
        :param clokta_config: the configuration containing okta connection information
        :type clokta_config: CloktaConfiguration
        :return: SUCCESS if succesfully got SAML token or INPUT_ERROR if cookie expired.
        Any other problem thows an exception.
        """
        self.saml_assertion = None
        result = self.__request_saml_assertion(configuration=clokta_config, use_session_token=False)
        return result

    def initiate_with_auth(self, clokta_config, mfas_to_fill):
        """
        Start the multistep process of getting a SAML token from Okta.  After this you need to
        call initiate_mfa
        :param clokta_config: clokta configuration with okta connection information
        :type clokta_config: CloktaConfiguration
        :param mfas_to_fill: an empty list that this will fill with possible MFA mechanisms
        :type mfas_to_fill: List[dict]
        :return: SUCCESS if successfully obtained SAML token, NEED_MFA if successfully identified user
        and retrieved MFA options, or FAIL if could not identify user because of bad password
        Any other result raises an exception
        """
        self.saml_assertion = None
        result = self.__auth_with_okta(
            configuration=clokta_config
        )

        # If no MFA was needed, then we've authed and can request SAML token
        if result == OktaInitiator.Result.SUCCESS:
            self.__request_saml_assertion(configuration=clokta_config, use_session_token=True)
        elif result == OktaInitiator.Result.NEED_MFA:
            mfas_to_fill[:] = self.factors
        return result

    def initiate_mfa(self, factor):
        """
        Next step after initiate_with_auth in the multistep process of getting a SAML token from Okta.
        Some MFA mechanism, like SMS, need to perform an action before prompting the user for response.
        After this state will still be NEED_MFA and you need to call finalize_mfa
        :param factor: the MFA mechanism to use
        :type factor: dict
        :return: whether a one time password will be needed to complete MFA
        :rtype: bool
        """

        if factor['factorType'] == 'sms':
            self.__okta_mfa_verification(
                factor_dict=factor,
                state_token=self.intermediate_state_token,
                otp_value=None
            )

        need_otp = factor['factorType'] != 'push'
        return need_otp

    def finalize_mfa(self, clokta_config, factor, otp):
        """
        Final step in multistep process of getting a SAML token from Okta.
        Submit MFA response to Okta
        :param clokta_config: clokta configuration with mfa response
        :type clokta_config: CloktaConfiguration
        :param factor: the MFA mechanism to use
        :type factor: dict
        :param otp: the one time password for MFA
        :type otp: str
        :return: SUCCESS if succesfully authenticated.  INPUT_ERROR if the otp was not correct.
        """

        if factor['factorType'] == 'push':
            result = self.__do_mfa_with_push(
                factor=factor,
                state_token=self.intermediate_state_token
            )
        else:
            result = self.__submit_mfa_response(factor=factor, otp=otp)

        if result == OktaInitiator.Result.SUCCESS:
            if Common.is_debug():
                Common.dump_out(message='Okta session token: {}'.format(self.session_token))

            # Now that we have a session token, request the SAML token
            self.__request_saml_assertion(configuration=clokta_config, use_session_token=True)
        return result

    def __request_saml_assertion(self, configuration, use_session_token):
        """
        request saml 2.0 assertion
        :param configuration: the clokta configuration
        :type configuration: CloktaConfiguration
        :param use_session_token: whether to submit a session token with the request.
        The session token will be obtained from self.session_token
        :type use_session_token: bool
        :return: SUCCESS if succesfully got SAML token or INPUT_ERROR if cookie expired.
        Any other problem thows an exception.
        """
        self.saml_assertion = None
        if use_session_token and not self.session_token:
            raise ValueError("No session token to use")

        response = self.__post_saml_request(
            use_session_token=use_session_token,
            configuration=configuration
        )

        if Common.is_debug():
            Common.dump_out(
                message='Requested SAML assertion from Okta {} session token.\nResponse: {}'.format(
                    "with" if use_session_token else "without", response.content
                )
            )

        soup = BeautifulSoup(response.content, "html.parser")
        for inputtag in soup.find_all('input'):
            if inputtag.get('name') == 'SAMLResponse':
                self.saml_assertion = inputtag.get('value')

        if not self.saml_assertion:
            if not use_session_token:
                # If a session token is not passed in, we consider a failure as a normal possibility
                if Common.is_debug():
                    Common.dump_out('Request without session token rejected.')
                return OktaInitiator.Result.INPUT_ERROR
            else:
                if Common.is_debug():
                    Common.dump_out(
                        'Expecting \'<input name="SAMLResponse" value="...">\' in Okta response, but not found.'
                    )
                raise RuntimeError('Unexpected response from Okta.')
        else:
            return OktaInitiator.Result.SUCCESS

    def __post_saml_request(self, use_session_token, configuration):
        """
        Send request to SAML token to Okta
        :param use_session_token: whether to pass a session token gotten from authenticating with Okta
        :type use_session_token: bool
        :param configuration: the clokta configuration with 'okta_aws_app_url'
        :type configuration: CloktaConfiguration
        :return: the HTTP response
        :rtype: json str
        """
        cookie_file = self.data_dir + 'clokta.cookies'
        url = configuration.get('okta_aws_app_url')
        if use_session_token:
            url += '?onetimetoken=' + self.session_token
        try:
            with open(cookie_file, 'rb') as f:
                cookies = pickle.load(f)
        except Exception:
            cookies = None
        response = requests.get(url, cookies=cookies)

        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            if not cookies:
                cookies = response.cookies
            else:
                cookies.update(response.cookies)
            with open(cookie_file, 'wb') as f:
                pickle.dump(cookies, f)
                return response
        else:
            response.raise_for_status()

    def __auth_with_okta(self, configuration):
        """
        Authenticate with Okta.  If no further info is required, return SUCCESS with session_token set.
        If MFA response is required, return NEED_MFA with intermediate_state_token and factors set
        A rejection from Okta is interpretted as a bad password and INPUT_ERROR is returned.
        :param configuration: clokta configuration containing connection and user info
        :type configuration: CloktaConfiguration
        :return: SUCCESS, NEED_MFA or INPUT_ERROR
        :rtype: OktaInitiator.Result
        """
        self.session_token = None
        self.factors = []

        try:
            okta_response = self.__post_auth_request(configuration)
        except requests.exceptions.HTTPError as http_err:
            if Common.is_debug():
                Common.dump_out(
                    ('Okta returned this credentials/password related error: {}\n' +
                        'This could be a mistyped password or a misconfigured username ' +
                        'or URL.').format(http_err))
            return OktaInitiator.Result.INPUT_ERROR
        except Exception as err:
            Common.dump_err('Unexpected error authenticating with Okta: {}'.format(err))
            raise

        if 'sessionToken' in okta_response and okta_response['sessionToken']:
            # MFA wasn't required.  We've got the token.
            self.session_token = okta_response['sessionToken']
            return OktaInitiator.Result.SUCCESS
        elif 'status' in okta_response and okta_response['status'] == 'MFA_ENROLL':
            # handle case where MFA is required but no factors have been enabled
            Common.dump_err('Please enroll in multi-factor authentication before using this tool')
            raise ValueError("No MFA mechanisms configured")
        elif 'status' in okta_response and okta_response['status'] == 'MFA_REQUIRED':
            self.factors = okta_response['_embedded']['factors']
            if not self.factors:
                # Another case where no factors have been enabled
                raise ValueError("No MFA mechanisms configured")
            self.intermediate_state_token = okta_response['stateToken']
            return OktaInitiator.Result.NEED_MFA
        else:
            Common.dump_err('Unexpected response from Okta authentication request')
            raise RuntimeError("Unexpected response from Okta")

    def __post_auth_request(self, configuration):
        """
        Posts a credentials-based authentication to Okta and returns an HTTP response
        :param configuration: clokta configuration with okta connection info
        :type configuration: CloktaConfiguration
        :return: the text from the HTTP response as a json blob
        :rtype: dict
        """
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cache-Control': '"no-cache',
            'Authorization': 'API_TOKEN'
        }
        payload = {
            'username': configuration.get('okta_username'),
            'password': configuration.get('okta_password')
        }
        org = self.__deduce_org(configuration.get('okta_aws_app_url'))
        url = 'https://{}/api/v1/authn'.format(org)

        response = requests.post(url, data=json.dumps(payload), headers=headers)
        if Common.is_debug():
            Common.dump_out(
                'Requested password-based authentication with Okta.\n' +
                'Response: {}'.format(response.content))

        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            resp = json.loads(response.text)
            return resp
        else:
            response.raise_for_status()

    def __deduce_org(self, app_url):
        """
        Pull the okta org (e.g. mycompany.okta.com) from the Okta app URL.
        :param app_url: the URL to an okta app (e.g. https://mycompany.okta.com/home/amazon_aws/hd63h3/542)
        :type app_url: str
        :return: the org's dns name (e.g. mycompany.okta.com)
        :rtype: str
        """
        start = len('https://')
        end = app_url.index('/', start+1)
        org = app_url[start:end]
        return org

    def __wait_for_push_result(self, state_token, push_response):
        """
        A request was sent to Okta querying the status of a push.  Process the response, pull the session
        token from it, and store in self.session_token
        :param state_token: a token received from Okta identifying this authentication attempt session
        :type state_token: str
        :param push_response: the HTTP response from the HTTP request
        :type push_response: json
        :return: SUCCESS if response indicates SUCCESS.  INPUT_ERROR if timed out.
        """
        url = push_response['_links']['next']['href']
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cache-Control': '"no-cache'
        }
        payload = {
            'stateToken': state_token
        }

        wait_for = 60
        timeout = time.time() + wait_for
        response_data = None
        while True:
            Common.echo(message='.', new_line=False)
            response = requests.post(url, data=json.dumps(payload), headers=headers)
            if response.status_code == requests.codes.ok:  # pylint: disable=E1101
                response_data = response.json()
            else:
                response.raise_for_status()

            if 'sessionToken' in response_data or time.time() > timeout:
                break
            time.sleep(3)

        if response_data and 'sessionToken' in response_data:
            Common.echo(message='Session confirmed')
            self.session_token = response_data['sessionToken']
            return OktaInitiator.Result.SUCCESS
        else:
            msg = 'Timeout expired ({} seconds)'.format(wait_for)
            Common.dump_err(message=msg)
            return OktaInitiator.Result.INPUT_ERROR

    def __okta_mfa_verification(self, factor_dict, state_token, otp_value=None):
        """Sends the MFA token entered and retuns the response"""
        url = factor_dict['_links']['verify']['href']
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cache-Control': '"no-cache'
        }
        payload = {
            'stateToken': state_token
        }
        if otp_value:
            payload['answer'] = otp_value

        data = json.dumps(payload)
        if Common.is_debug():
            Common.dump_out("Sending MFA verification to...\nurl: {}\nbody: {}".format(url, data))
        response = requests.post(url, data=data, headers=headers)
        if Common.is_debug():
            Common.dump_out(
                "Received {} response from Okta: {}".format(response.status_code, json.dumps(response.json()))
            )
        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            return response.json()
        else:
            response.raise_for_status()

    def __submit_mfa_response(self, factor, otp):
        """
        post one time password to Okta
        :param factor: the MFA mechanism to use
        :type factor: dict
        :param otp: the one time password for MFA
        :type otp: str
        :return: SUCCESS if received a session token from Okta.  INPUT_ERROR if Okta reports incorrect
        one time password
        :rtype: OktaInitiator.Result
        """
        self.session_token = None
        try:
            mfa_response = self.__okta_mfa_verification(
                factor_dict=factor,
                state_token=self.intermediate_state_token,
                otp_value=otp
            )
            if 'sessionToken' in mfa_response:
                self.session_token = mfa_response['sessionToken']
                return OktaInitiator.Result.SUCCESS
            else:
                if 'status' not in mfa_response:
                        raise ValueError('Unexpected response from Okta.  ' +
                                         'Received neither session token nor error status.')
                Common.dump_err('Okta failed to return session token and only returned this status: {}'.format(
                        mfa_response['status']))
                return OktaInitiator.Result.INPUT_ERROR
        except requests.exceptions.HTTPError as http_err:
            Common.dump_err('Okta returned this MFA related error: {}'.format(http_err))
            return OktaInitiator.Result.INPUT_ERROR
        except Exception as err:
            msg = 'Unexpected error: {}'.format(err)
            Common.dump_err(message=msg)
            raise ValueError("Unexpected error with MFA")

    def __do_mfa_with_push(self, factor, state_token):
        """
        Send push re: Okta Verify and wait for response.
        If succesful, session token will be stored in self.session_token
        :param factor: mfa information
        :type factor: dict
        :param state_token: token used in MFA back and forth
        :type: str
        :return: SUCCESS if push reported success.  INPUT_ERROR if user never responded.  Any other
        possibilities will result in an exception
        :rtype: OktaInitiator.Result
        """
        url = factor['_links']['verify']['href']
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Cache-Control': '"no-cache'
        }
        payload = {
            'stateToken': state_token
        }

        response_data = None
        response = requests.post(url, data=json.dumps(payload), headers=headers)
        if response.status_code == requests.codes.ok:  # pylint: disable=E1101
            response_data = response.json()
        else:
            response.raise_for_status()

        Common.echo(message='Push notification sent; waiting for your response', new_line=False)

        status = response_data['status']
        if status == 'MFA_CHALLENGE':
            if 'factorResult' in response_data and response_data['factorResult'] == 'WAITING':
                return self.__wait_for_push_result(
                    state_token=state_token,
                    push_response=response_data
                )




''' RoleChooser class must be instantiated prior to use '''

import click

from clokta.common import Common


class RoleChooser(object):
    """
    Supports AWS Role determination
    """

    def __init__(self, possible_roles, role_preference=None):
        """
        :param possible_roles: list of possible roles to choose from
        :type possible_roles: List[AwsRole]
        :param role_preference: preferred role
        :type role_preference: str
        """
        self.possible_roles = possible_roles
        self.role_preference = role_preference

    def choose_role(self):
        """
        Look for a default role defined and, if not, prompt the user for one
        Allow the user to also specify the role is the default to be used
        from now on
        :return: a tuple of the chosen role and whether it is the
        new default
        :rtype: AwsRole, bool
        """
        # throw an error if no roles are provided
        #  (defensive coding only - this should not be possible)
        if not self.possible_roles:
            Common.dump_err(
                message='No AWS Role was assigned to this application!'
            )
            raise ValueError('Unexpected configuration - No AWS role assigned to Okta login.')

        # use the one provided if there is only one
        if len(self.possible_roles) == 1:
            role = self.possible_roles[0]
            if self.role_preference and role.role_arn != self.role_preference:
                Common.dump_err(
                    message='Your cofigured role "{notfound}" was not found; using "{found}" role'.format(
                        notfound=self.role_preference,
                        found=role.role_name
                    )
                )
            elif Common.is_debug():
                Common.echo(
                    message="Using default role '{role}'".format(
                        role=role.role_arn
                    )
                )
            return role, False

        # use the configured role if it matches one from the the SAML assertion
        for role in self.possible_roles:
            if role.role_arn == self.role_preference:
                message = "Using default role '{}'".format(role.role_name)
                extra_message = '.  Run "clokta --no-default-role" to override.'
                if Common.get_output_format() == Common.long_out:
                    Common.echo(message + extra_message)
                else:
                    Common.echo(message)
                return role, True

        # make the user choose
        return self.__prompt_for_role(with_set_default_option=True)

    def __prompt_for_role(self, with_set_default_option):
        """
        Give the user a choice from the intersection of configured and supported factors
        :param with_set_default_option: if True will add an option for setting a default role
        :type with_set_default_option: bool
        :return: a tuple of what role was chosen and whether it is the new default
        :rtype: AwsRole, bool
        """
        index = 1
        for role in self.possible_roles:
            msg = '{index} - {prompt}'.format(
                index=index,
                prompt=role.role_name
            )
            Common.echo(message=msg, bold=True)
            index += 1
        if with_set_default_option:
            Common.echo('{index} - set a default role'.format(index=index))
        raw_choice = None
        try:
            raw_choice = click.prompt(text='Choose a Role ARN to use', type=int, err=Common.to_std_error())
            choice = raw_choice - 1
        except ValueError:
            Common.echo(message='Please select a valid option: you chose: {}'.format(raw_choice))
            return self.__prompt_for_role()

        if choice == len(self.possible_roles):
            # They want to set a default.  Prompt again (just without the set-default option)
            # and return that chosen role and that it's the new default
            chosen_option, _ = self.__prompt_for_role(with_set_default_option=False)
            return chosen_option, True
        if len(self.possible_roles) > choice >= 0:
            pass
        else:
            Common.echo(message='Please select a valid option: you chose: {}'.format(raw_choice))
            return self.__prompt_for_role(with_set_default_option=with_set_default_option)

        chosen_option = self.possible_roles[choice]
        if Common.is_debug():
            Common.dump_out(
                message='Using chosen Role {role} & IDP {idp}'.format(
                    role=chosen_option.role_arn,
                    idp=chosen_option.idp_arn
                )
            )

        return chosen_option, False



import base64
import os

import boto3
from botocore.exceptions import ClientError
import xml.etree.ElementTree as ElementTree

from clokta.clokta_configuration import CloktaConfiguration
from clokta.common import Common
from clokta.awsrole import AwsRole


class AwsCredentialsGenerator:
    """
    This interfaces with AWS and, given a Okta-generated SAML token, can
    generate AWS credentials for a given role
    """

    def __init__(self, data_dir, clokta_config, saml_assertion):
        """
        Creates a credential generator capable of creating credentials in AWS
        :param clokta_config: the clokta configuration with which role has been designated default
        :type clokta_config: CloktaConfiguration
        :param saml_assertion: the saml token generated by Okta.  This will define what account and roles will be used
        :type saml_assertion: str
        """
        self.clokta_config = clokta_config
        self.saml_assertion = saml_assertion
        self.data_dir = data_dir
        self.bash_file = None  # type: str
        self.docker_file = None  # type: str
        self.roles = self.__deduce_roles_from_saml()  # type: [AwsRole]

        # We need to make sure, when interacting with AWS, we don't try to use
        # default creds as we are creating our own.
        envars_to_clear = [
            "AWS_PROFILE", 
            "AWS_ACCESS_KEY_ID",
            "AWS_SECRET_ACCESS_KEY",
            "AWS_SESSION_TOKEN"
        ]
        for envvar in envars_to_clear:
            if envvar in os.environ:
                del os.environ[envvar]


    def get_roles(self):
        """
        Returns the possible roles that can be assumed with the SAML token
        :return: list of roles
        :rtype: [AwsRole]
        """
        return self.roles

    def generate_creds(self, role):
        """
        :param role: the AWS role the user wants to assume
        :type role: AwsRole
        """

        client = boto3.client('sts')
        # Try for a 12 hour session.  If it fails, try for shorter periods
        assumed_role_credentials = None
        durations = [43200, 14400, 3600]
        for duration in durations:
            try:
                assumed_role_credentials = client.assume_role_with_saml(
                    RoleArn=role.role_arn,
                    PrincipalArn=role.idp_arn,
                    SAMLAssertion=self.saml_assertion,
                    DurationSeconds=duration
                )
                if duration == 3600:
                    Common.echo(message='YOUR SESSION WILL ONLY LAST ONE HOUR')
                break
            except ClientError as e:
                # If we get a validation error and we have shorter durations to try, try a shorter duration
                if e.response['Error']['Code'] != 'ValidationError' or duration == durations[-1]:
                    raise

        self.clokta_config.apply_credentials(credentials=assumed_role_credentials)
        self.bash_file = self.__write_sourceable_file(credentials=assumed_role_credentials)
        self.docker_file = self.__write_dockerenv_file(credentials=assumed_role_credentials)

    def __deduce_roles_from_saml(self):
        """
        Parse the SAML Token which contains the roles that can be assumed
        :return: roles contained in SAML token
        :rtype: [Role]
        """
        roles = []
        decoded_assertion = base64.b64decode(self.saml_assertion)
        root = ElementTree.fromstring(decoded_assertion)
        for saml2_attribute in root.iter('{urn:oasis:names:tc:SAML:2.0:assertion}Attribute'):
            if saml2_attribute.get('Name') == 'https://aws.amazon.com/SAML/Attributes/Role':
                for saml2_attr_value in saml2_attribute.iter('{urn:oasis:names:tc:SAML:2.0:assertion}AttributeValue'):
                    next_role = AwsRole(saml2_attr_value.text)
                    roles.append(next_role)
        return roles

    def __write_sourceable_file(self, credentials):
        """
        Generates a shell script to source in order to apply credentials to the shell environment.
        """
        creds = credentials['Credentials']
        output_file_name = '{dir}{profile}.sh'.format(
            dir=os.path.expanduser(self.data_dir),
            profile=self.clokta_config.profile_name
        )
        lines = [
            'export AWS_ACCESS_KEY_ID={}\n'.format(creds['AccessKeyId']),
            'export AWS_SECRET_ACCESS_KEY={}\n'.format(creds['SecretAccessKey'])
        ]
        if 'SessionToken' in creds:
            lines.append('export AWS_SESSION_TOKEN={}\n'.format(creds['SessionToken']))
        else:
            lines.append('unset AWS_SESSION_TOKEN')

        with open(output_file_name, mode='w') as file_handle:
            file_handle.writelines(lines)

        short_output_file_name = '{dir}{profile}.sh'.format(
            dir=self.data_dir,
            profile=self.clokta_config.profile_name
        )
        return short_output_file_name

    def __write_dockerenv_file(self, credentials):
        """
        Generates a Docker .env file that can be used with docker compose to inject into the environment.
        """
        creds = credentials['Credentials']
        output_file_name = '{dir}{profile}.env'.format(
            dir=os.path.expanduser(self.data_dir),
            profile=self.clokta_config.profile_name
        )
        lines = [
            'AWS_ACCESS_KEY_ID={}\n'.format(creds['AccessKeyId']),
            'AWS_SECRET_ACCESS_KEY={}\n'.format(creds['SecretAccessKey'])
        ]
        if 'SessionToken' in creds:
            lines.append('AWS_SESSION_TOKEN={}\n'.format(creds['SessionToken']))

        with open(output_file_name, mode='w') as file_handle:
            file_handle.writelines(lines)

        short_output_file_name = '{dir}{profile}.env'.format(
            dir=self.data_dir,
            profile=self.clokta_config.profile_name
        )
        return short_output_file_name




class Factors(object):

    @staticmethod
    def mfa_providers():
        return [
            {
                'id': 'GoogleAuthenticator',
                'prompt': 'Google Authenticator',
                'factor_type': 'token:software:totp',
                'provider': 'GOOGLE'
            }, {
                'id': 'OktaVerify',
                'prompt': 'Okta Verify',
                'factor_type': 'token:software:totp',
                'provider': 'OKTA'
            }, {
                'id': 'Sms',
                'prompt': 'SMS text message',
                'factor_type': 'sms',
                'provider': 'OKTA'
            }, {
                'id': 'OktaPush',
                'prompt': 'Okta Verify with Push',
                'factor_type': 'push',
                'provider': 'OKTA'
            }
        ]



''' Define executable setup '''
import os
from setuptools import setup, find_packages
import warnings

setup(
    name='clokta',
    version='4.1.3-alpha.1',
    packages=find_packages(),
    include_package_data=True,
    py_modules=['clokta'],
    install_requires=[
        'beautifulsoup4',
        'boto3',
        'click>=7.0',
        'configparser',
        'enum-compat',
        'keyring',
        'requests'
    ],
    entry_points={
        'console_scripts': [
            'clokta=clokta.cloktacli:assume_role'
        ]
    },
    namespace_packages=['clokta'],
    author="Robert Antonucci and the WaPo platform tools team",
    author_email="opensource@washingtonpost.com",
    url="https://github.com/washingtonpost/clokta",
    download_url="https://github.com/washingtonpost/clokta/tarball/4.1.3-alpha.1",
    keywords=['okta', 'clokta', 'aws', 'cli'],
    classifiers=[]
)



class AwsRole:
    """Represents an AWS role to generate credentials for"""
    def __init__(self, idp_role_str):
        """
        :param idp_role_str: a single comma-separated string with the ARN of the Okta IdP and the ARN of the role
        :type idp_role_str: str
        """
        split_str = idp_role_str.split(',')
        self.idp_arn = split_str[0]
        self.role_arn = split_str[1]
        self.account = self.role_arn.split(':')[4]  # type: str
        self.role_name = self.role_arn.split(':')[5][5:]  # type: str



'''
Simple utility methods for the module
'''
import logging
import sys
from datetime import date, datetime

import click


class Common(object):

    quiet_out = 0
    brief_out = 1
    long_out = 2
    debugging_out=3
    output_format=brief_out

    ''' Console printing helpers '''

    @classmethod
    def get_output_format(cls):
        return Common.output_format

    @classmethod
    def set_output_format(cls, new_format):
        Common.output_format=new_format

    @classmethod
    def is_debug(cls):
        """
        :return: whether to output debugging information
        :rtype: bool
        """
        return Common.output_format == Common.debugging_out

    @classmethod
    def to_std_error(cls):
        """
        :return: whether clokta is running in a special mode for piping unix commands where most normal output is actually
        sent to stderr and only executable statements are output to stdout
        :rtype: bool
        """
        return Common.output_format == Common.quiet_out

    @classmethod
    def dump_err(cls, message, new_line=True):
        """
        Print error information to the user
        :param message: the message to print
        :param new_line: whether to put a new line at the end (default is include new line)
        """
        click.secho(message, nl=new_line, bold=True, fg='red', err=True)

    @classmethod
    def dump_out(cls, message, new_line=True):
        """
        Print debug information to the user
        :param message: the message to print
        :param new_line: whether to put a new line at the end (default is include new line)
        """
        if new_line:
            message += '\n'
        if Common.is_debug():
            click.secho(message, nl=new_line, bold=False, fg='blue', err=Common.to_std_error())

    @classmethod
    def echo(cls, message, new_line=True, bold=False, always_stdout=False):
        """
        Print message to user.  This is meant for info level messages and prompts for information
        :param message: message to print
        :param new_line: whether to put a new line at the end (default is include new line)
        :param bold: whether to make the text bold
        :param always_stdout: sometimes users will pipe the output of clokta to another shell script and
        all prompting and info messages should be sent to stderr and only executable commands should be sent to
        stdout.  always_stdout should be True if this output is an executable command that should go to stdout then
        """
        to_std_error = not always_stdout and Common.to_std_error()
        click.secho(message, nl=new_line, bold=bold, err=to_std_error)

    @classmethod
    def json_serial(cls, obj):
        ''' JSON serializer for objects not serializable by default json code '''
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        raise TypeError("Type %s not serializable" % type(obj))



"""
This is the entry-point to the cli application.
"""
import os

import click

from clokta.clokta_configuration import CloktaConfiguration
from clokta.role_assumer import RoleAssumer, Common


@click.command()
@click.version_option()
@click.option('--profile', '-p', help='Configuration profile.  Required unless specified by AWS_PROFILE')
@click.option('--inline-help', '-i', is_flag=True,
              help='Output explicit steps on how to use generated keys and override defaults')
@click.option('--no-default-role', is_flag=True, help='Lets you choose a different role than your default')
@click.option('--quiet', '-q', is_flag=True,
              help='Silences all output except final export command. All prompts are on stderr. ' +
                   'This facilitate commands like "eval $(clokta -p default)"')
@click.option('--verbose', '-v', is_flag=True, help='Output internal state for debugging')
@click.option('--list-accounts',  is_flag=True,
              help='List all accounts, profile and account number, configured in clokta')
def assume_role(profile, inline_help=False, no_default_role=False, quiet=False, verbose=False, list_accounts=False):
    """ Click point of entry """

    if list_accounts:
        CloktaConfiguration.dump_account_numbers('~/.clokta/clokta.cfg')
        exit(0)

    if not profile:
        profile = get_profile_from_env()
        if not profile:
            with click.Context(assume_role) as ctx:
                click.echo(assume_role.get_help(ctx))
                exit(0)

    configure_output_format(verbose, inline_help, quiet)
    assumer = RoleAssumer(profile=profile)
    assumer.assume_role(reset_default_role=no_default_role)


def configure_output_format(verbose, inline_help, quiet):
    """
    Reads the three output-related command line flags and determines desired output 
    """
    if verbose:
        Common.set_output_format(Common.debugging_out)
    elif quiet:
        Common.set_output_format(Common.quiet_out)
    elif inline_help:
        Common.set_output_format(Common.long_out)
    else:
        Common.set_output_format(Common.brief_out)


def get_profile_from_env():
    """
    Look up the AWS_PROFILE variable and return it
    :return: value of AWS_PROFILE, or None if not defined
    :rtype: stre
    """
    from_env = os.getenv(key='AWS_PROFILE')
    if from_env:
        Common.echo("Using profile '{}' from AWS_PROFILE".format(from_env))
    return from_env
